IMPLEMENTATION MODULE Conversions;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     RcsId: $Id: conversi.mod,v 1.4 92/07/27 20:47:15 paul Rel $

  *)

  (*$O-*)

  FROM Strings IMPORT
    Assign, Length, Insert, Reverse, CharIndex;

  (* ------------------------------------------------------------------
                   Implementation of AUXILIARY procedures
     ------------------------------------------------------------------ *)

  PROCEDURE Number(    numeral: CHAR;
                   VAR number:  CARDINAL;
                   VAR done:    BOOLEAN);
  BEGIN
    (* initialise *)
    done := TRUE;
    numeral := CAP(numeral);

    (* translate numeral to number *)
    IF ('0' <= numeral) & (numeral <= '9') THEN
      number := ORD(numeral) - ORD('0')
    ELSIF ('A' <= numeral) & (numeral <= 'Z') THEN
      number := ORD(numeral) - (ORD('A') - 10)
    ELSE
      done := FALSE
    END
  END Number;

  (* ------------------------------------------------------------------
                    Implementation of DEFINED procedures
     ------------------------------------------------------------------ *)

  PROCEDURE Numeral(num: CARDINAL): CHAR;
    VAR ch: CHAR;
  BEGIN
    IF num < 10 THEN ch := CHR(num + ORD('0'))
    ELSE ch := CHR(num + (ORD('A') - 10))
    END;
    RETURN ch
  END Numeral;

  PROCEDURE Numeral10(num: CARDINAL): CHAR;
  BEGIN
    RETURN Numeral(num MOD 10)
  END Numeral10;

  PROCEDURE ConvertToString(    num:  LONGCARD;
                                len:  CARDINAL;
                                base: CARDINAL;
                                sign: SignType;
                            VAR str:  ARRAY OF CHAR);
    VAR strbuf: ARRAY [0..33] OF CHAR;
        stridx: CARDINAL;
        dbase: LONGCARD;

    PROCEDURE Char(ch: CHAR);
    BEGIN
      IF ch # 0C THEN
        strbuf[stridx] := ch; INC(stridx)
      END
    END Char;

  BEGIN
    (* truncate len to maximum possible length of string *)
    IF len > HIGH(str)+1 THEN len := HIGH(str)+1 END;

    (* initialise *)
    dbase := VAL(LONGCARD, base);
    stridx := 0;

    (* convert the number; digits are in reverse order *)
    REPEAT
      Char(Numeral(VAL(CARDINAL, num MOD dbase)));
      num := num DIV dbase
    UNTIL num = 0;

    (* append the sign and terminate the string *)
    Char(CharIndex('+- ', ORD(sign)));
    strbuf[stridx] := 0C;

    (* reverse buffer *)
    Reverse(strbuf);
    Assign(strbuf, str);

    (* right-justify string *)
    WHILE stridx < len DO
      Insert(' ', str, 0); INC(stridx)
    END
  END ConvertToString;

  (*$V-*)
  PROCEDURE ConvertToNumber(     str:   ARRAY OF CHAR;
                                 base:  CARDINAL;
                                 negok: BOOLEAN;
                                 max:   LONGCARD;
                             VAR num:   LONGCARD;
                             VAR done:  BOOLEAN);
  VAR negate: BOOLEAN;
      dbase: LONGCARD;
      i, len, digit, mmb: CARDINAL;
      mdb: LONGCARD;
  BEGIN
    done := (base >= 2) & (base <= 36);
    IF done THEN
      dbase := VAL(LONGCARD, base);
      mmb := VAL(CARDINAL, max MOD dbase);
      mdb := max DIV dbase;
      i := 0; num := 0;
      len := Length(str);
      negate := FALSE;
      IF str[0] = '+' THEN INC(i)
      ELSIF negok & (str[0] = '-') THEN INC(i); negate := TRUE
      END;
      done := len > i;
      WHILE (i < len) & done DO
        Number(str[i], digit, done);
        done := done & (digit < base);
        IF done THEN
          done := (num < mdb) OR ((num = mdb) & (digit <= mmb));
          IF done THEN
            num := num * dbase + VAL(LONGCARD, digit);
            INC(i)
          END
        END
      END;
      IF negate THEN num := LONGCARD(-LONGINT(num)) END
    END
  END ConvertToNumber;
  (*$V=*)

  PROCEDURE ConvertOctal(    num: CARDINAL;
                             len: CARDINAL;
                         VAR str: ARRAY OF CHAR);
  BEGIN
    ConvertToString(VAL(LONGCARD, num), len, Oct, none, str)
  END ConvertOctal;

  PROCEDURE ConvertHex(    num: CARDINAL;
                           len: CARDINAL;
                       VAR str: ARRAY OF CHAR);
  BEGIN
    ConvertToString(VAL(LONGCARD, num), len, Hex, none, str)
  END ConvertHex;

  PROCEDURE ConvertCardinal(    num: CARDINAL;
                                len: CARDINAL;
                            VAR str: ARRAY OF CHAR);
  BEGIN
    ConvertToString(VAL(LONGCARD, num), len, Dec, none, str)
  END ConvertCardinal;

  PROCEDURE ConvertInteger(    num: INTEGER;
                               len: CARDINAL;
                           VAR str: ARRAY OF CHAR);
    VAR sign: SignType;
        n: LONGCARD;
  BEGIN
    IF num < 0 THEN sign := minus
    ELSE sign := none
    END;
    IF num = MIN(INTEGER) THEN
      (* In two's complement arithmetic, MIN(INTEGER) is a special case *)
      n := ABS(MIN(INTEGER))
    ELSE
      n := VAL(LONGCARD, ABS(num))
    END;
    ConvertToString(n, len, Dec, sign, str)
  END ConvertInteger;

  PROCEDURE ConvertLongOctal(    num: LONGCARD;
                                 len: CARDINAL;
                             VAR str: ARRAY OF CHAR);
  BEGIN
    ConvertToString(num, len, Oct, none, str)
  END ConvertLongOctal;

  PROCEDURE ConvertLongHex(    num: LONGCARD;
                               len: CARDINAL;
                           VAR str: ARRAY OF CHAR);
  BEGIN
    ConvertToString(num, len, Hex, none, str)
  END ConvertLongHex;

  PROCEDURE ConvertLongcard(    num: LONGCARD;
                                len: CARDINAL;
                            VAR str: ARRAY OF CHAR);
  BEGIN
    ConvertToString(num, len, Dec, none, str)
  END ConvertLongcard;

  PROCEDURE ConvertLongint(    num: LONGINT;
                               len: CARDINAL;
                           VAR str: ARRAY OF CHAR);
    VAR sign: SignType;
  BEGIN
    IF num < 0 THEN sign := minus
    ELSE sign := none
    END;
    ConvertToString(ABS(num), len, Dec, sign, str)
  END ConvertLongint;

  (*$V-*)
  PROCEDURE DecimalToCardinal(    str:  ARRAY OF CHAR;
                              VAR num:  CARDINAL;
                              VAR done: BOOLEAN);
    VAR n: LONGCARD;
  BEGIN
    ConvertToNumber(str, Dec, FALSE, MAX(CARDINAL), n, done);
    IF done THEN
      num := VAL(CARDINAL, n)
    END
  END DecimalToCardinal;
  (*$V=*)

  (*$V-*)
  PROCEDURE HexToCardinal(    str:  ARRAY OF CHAR;
                          VAR num:  CARDINAL;
                          VAR done: BOOLEAN);
    VAR n: LONGCARD;
  BEGIN
    ConvertToNumber(str, Hex, FALSE, MAX(CARDINAL), n, done);
    IF done THEN
      num := VAL(CARDINAL, n)
    END
  END HexToCardinal;
  (*$V=*)

  (*$V-*)
  PROCEDURE OctalToCardinal(    str:  ARRAY OF CHAR;
                            VAR num:  CARDINAL;
                            VAR done: BOOLEAN);
    VAR n: LONGCARD;
  BEGIN
    ConvertToNumber(str, Oct, FALSE, MAX(CARDINAL), n, done);
    IF done THEN
      num := VAL(CARDINAL, n)
    END
  END OctalToCardinal;
  (*$V=*)

  (*$V-*)
  PROCEDURE DecimalToInteger(    str:  ARRAY OF CHAR;
                             VAR num:  INTEGER;
                             VAR done: BOOLEAN);
    VAR n: LONGCARD;
  BEGIN
    ConvertToNumber(str, Dec, TRUE, MAX(INTEGER), n, done);
    IF done THEN
      num := VAL(INTEGER, LONGINT(n))
    END
  END DecimalToInteger;
  (*$V=*)

  (*$V-*)
  PROCEDURE DecimalToLongcard(    str:  ARRAY OF CHAR;
                              VAR num:  LONGCARD;
                              VAR done: BOOLEAN);
  BEGIN
    ConvertToNumber(str, Dec, FALSE, MAX(LONGCARD), num, done);
  END DecimalToLongcard;
  (*$V=*)

  (*$V-*)
  PROCEDURE HexToLongcard(    str:  ARRAY OF CHAR;
                          VAR num:  LONGCARD;
                          VAR done: BOOLEAN);
  BEGIN
    ConvertToNumber(str, Hex, FALSE, MAX(LONGCARD), num, done);
  END HexToLongcard;
  (*$V=*)

  (*$V-*)
  PROCEDURE OctalToLongcard(    str:  ARRAY OF CHAR;
                            VAR num:  LONGCARD;
                            VAR done: BOOLEAN);
  BEGIN
    ConvertToNumber(str, Oct, FALSE, MAX(LONGCARD), num, done);
  END OctalToLongcard;
  (*$V=*)

  (*$V-*)
  PROCEDURE DecimalToLongint(    str:  ARRAY OF CHAR;
                             VAR num:  LONGINT;
                             VAR done: BOOLEAN);
    VAR n: LONGCARD;
  BEGIN
    ConvertToNumber(str, Dec, TRUE, MAX(LONGINT), n, done);
    num := LONGINT(n);
  END DecimalToLongint;
  (*$V=*)

END Conversions.

