IMPLEMENTATION MODULE FileSystem;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     $Id: filesyst.mod,v 1.2 90/12/10 16:23:08 paul Exp $

  *)

  (*$V-*)

  FROM SYSTEM IMPORT
    ADR, ADDRESS, BYTE, WORD, LONGWORD, ADDADR;
  FROM Types IMPORT
    off_t;
  FROM SysLib IMPORT
    StreamPtr, Flags_MSdos;
  FROM POSIX IMPORT
    open, read, write, lseek, close, unlink, fdstream, SEEK_SET, errno;
  FROM Fcntl IMPORT
    O_CREAT, O_RDWR, O_TRUNC;
  FROM Storage IMPORT
    Allocate, Deallocate;
  FROM Strings IMPORT
    Length, Assign;
  FROM ASCII IMPORT
    EOF, EOL, CR, LF;

  CONST
    BufSize = 1024;

  TYPE
    FileStatus  = (Reading, Writing);
    Internal    = POINTER TO InternalRec;
    InternalRec = RECORD
                    bp, be:  INTEGER;
                    dirty:   BOOLEAN;
                    path:    INTEGER;
                    filePos: off_t;
                    OSEOL:   CHAR;
                    buffer:  ARRAY [0..BufSize-1] OF BYTE;
                  END;

  PROCEDURE FillBuffer(VAR f: File);
    VAR pos: off_t;
        err: INTEGER;
  BEGIN
    f.eof := FALSE;
    WITH f.int^ DO
      dirty := FALSE;
      pos := lseek(path, filePos, SEEK_SET);
      err := read(path, ADR(buffer), SIZE(buffer));
      IF err = -1 THEN
        f.err := errno; f.res := notdone
      ELSIF err = 0 THEN
        f.res := eom; f.eof := TRUE
      ELSE
        f.res := done
      END;
      bp := 0; be := err
    END
  END FillBuffer;

  PROCEDURE WriteBuffer(VAR f: File);
    VAR err: INTEGER;
        pos: off_t;
  BEGIN
    WITH f.int^ DO
      f.err := 0;
      IF dirty THEN
        pos := lseek(path, filePos, SEEK_SET);
        err := write(path, ADR(buffer), bp);
        IF err = -1 THEN
          f.err := errno; f.res := notdone
        ELSIF err # bp THEN
          f.res := eom; f.eof := TRUE
        ELSE
          f.res := done
        END;
        INC(filePos, bp); bp := 0; be := 0
      ELSE
        INC(filePos, bp); bp := 0; be := 0;
        pos := lseek(path, filePos, SEEK_SET);
      END;
      dirty := FALSE
    END
  END WriteBuffer;

  PROCEDURE OpenFile(VAR f: File; VAR name: ARRAY OF CHAR; newFile: BOOLEAN);
    VAR NTname: ARRAY [0..199] OF CHAR;
        err: INTEGER;
        strm: StreamPtr;
  BEGIN
    Assign(name, NTname);
    NTname[HIGH(NTname)] := 0C;
    WITH f.int^ DO
      f.new := newFile;
      IF newFile THEN
        err := unlink(NTname);
        err := open(NTname, O_CREAT+O_RDWR+O_TRUNC);
        IF err >= 0 THEN f.err := 0; path := err
        ELSE f.err := err
        END
      ELSE
        err := open(NTname, O_RDWR);
        IF err >= 0 THEN f.err := 0; path := err
        ELSE f.err := err
        END
      END;
      IF f.err = 0 THEN
        f.res := done;
        f.eof := FALSE;
        filePos := 0; bp := 0; be := 0; dirty := FALSE;
        strm := fdstream(path);
        IF Flags_MSdos IN BITSET(strm^.Flags) THEN OSEOL := CR
        ELSE OSEOL := LF
        END;
        IF NOT newFile THEN FillBuffer(f) END;
      ELSE
        IF newFile THEN f.res := notdone
        ELSE f.res := unknownfile
        END
      END
    END
  END OpenFile;

  PROCEDURE Lookup(VAR f: File; name: ARRAY OF CHAR; new: BOOLEAN);
  BEGIN
    Allocate(f.int, SIZE(f.int^));
    IF f.int = NIL THEN
      f.res := notdone
    ELSE
      OpenFile(f, name, FALSE);
      IF (f.res # done) & new THEN
        OpenFile(f, name, TRUE)
      END;
      IF f.res # done THEN
        Deallocate(f.int, SIZE(f.int^))
      END
    END
  END Lookup;

  PROCEDURE Create(VAR f: File; name: ARRAY OF CHAR);
  BEGIN
    Allocate(f.int, SIZE(f.int^));
    IF f.int = NIL THEN
      f.res := notdone
    ELSE
      OpenFile(f, name, TRUE);
      IF f.res # done THEN
        Deallocate(f.int, SIZE(f.int^))
      END
    END
  END Create;

  PROCEDURE Close(VAR f: File);
    VAR err: CARDINAL;
  BEGIN
    WriteBuffer(f);
    IF f.res = done THEN
      f.err := close(f.int^.path);
      IF f.err # 0 THEN f.res := notdone END
    ELSE
      err := close(f.int^.path)
    END;
    Deallocate(f.int, SIZE(f.int^))
  END Close;

  PROCEDURE Delete(name: ARRAY OF CHAR);
    VAR err: CARDINAL;
        NTname: ARRAY [0..199] OF CHAR;
  BEGIN
    Assign(name, NTname);
    NTname[HIGH(NTname)] := 0C;
    err := unlink(NTname)
  END Delete;

  PROCEDURE Rename(oldname: ARRAY OF CHAR;
                   newname: ARRAY OF CHAR);
  BEGIN
    (* not implemented *)
  END Rename;

  PROCEDURE GetPos(VAR f: File; VAR pos: LONGCARD);
  BEGIN
    pos := f.int^.filePos + LONGCARD(f.int^.bp);
  END GetPos;

  PROCEDURE SetPos(VAR f: File; pos: LONGCARD);
    VAR fpos: off_t;
  BEGIN
    WITH f.int^ DO
      f.err := 0;
      WriteBuffer(f);
      IF f.res = done THEN
        fpos := lseek(path, pos, SEEK_SET);
        IF INTEGER(fpos) = -1 THEN
          f.err := errno
        ELSE
          filePos := pos;
          bp := 0; be := 0;
          f.eof := FALSE
        END
      END;
      IF f.err = 0 THEN f.res := done
      ELSE f.res := notdone
      END
    END
  END SetPos;

  PROCEDURE Reset(VAR f: File);
  BEGIN
    SetPos(f, 0)
  END Reset;

  PROCEDURE ReadLn(VAR f: File);
    VAR c: CHAR;
  BEGIN
    ReadChar(f, c);
    WHILE (c # EOL) & NOT f.eof DO
      ReadChar(f, c)
    END
  END ReadLn;

  PROCEDURE ReadByte(VAR f: File; VAR b: BYTE);
  BEGIN
    IF f.eof THEN
      f.res := callerror; b := BYTE(0)
    ELSE
      WITH f.int^ DO
        IF bp = be THEN
          WriteBuffer(f);
          FillBuffer(f);
        END;
        b := buffer[bp]; INC(bp)
      END
    END
  END ReadByte;

  PROCEDURE ReadChar(VAR f: File; VAR ch: CHAR);
  BEGIN
    ReadByte(f, ch);
    IF ch = f.int^.OSEOL THEN
      IF ch = CR THEN ReadByte(f, ch) END; (* skip LF *)
      ch := EOL
    END
  END ReadChar;

  PROCEDURE WriteByte(VAR f: File; b: BYTE);
    VAR count: CARDINAL;
  BEGIN
    WITH f.int^ DO
      dirty := TRUE;
      f.err := 0;
      buffer[bp] := b;
      INC(bp);
      IF f.int^.bp >= BufSize THEN
        WriteBuffer(f)
      END
    END
  END WriteByte;

  PROCEDURE WriteChar(VAR f: File; ch: CHAR);
  BEGIN
    IF ch = EOL THEN
      IF f.int^.OSEOL = CR THEN
        WriteByte(f, CR); WriteByte(f, LF)
      ELSE
        WriteByte(f, LF)
      END
    ELSE WriteByte(f, ch)
    END
  END WriteChar;

  PROCEDURE ReadWord(VAR f: File; VAR w: WORD);
    VAR b0, b1: BYTE;
  BEGIN
    ReadByte(f, b0); ReadByte(f, b1);
    w := WORD(CARDINAL(b0)*100H + CARDINAL(b1))
  END ReadWord;

  PROCEDURE WriteWord(VAR f: File; w: WORD);
  BEGIN
    WriteByte(f, BYTE(CARDINAL(w) DIV 100H));
    WriteByte(f, BYTE(CARDINAL(w) MOD 100H));
  END WriteWord;

  PROCEDURE ReadLong(VAR f: File; VAR w: LONGWORD);
    VAR w0, w1: WORD;
  BEGIN
    ReadWord(f, w0); ReadWord(f, w1);
    w := LONGWORD(LONGCARD(w0)*10000H + LONGCARD(w1))
  END ReadLong;

  PROCEDURE WriteLong(VAR f: File; w: LONGWORD);
  BEGIN
    WriteWord(f, WORD(LONGCARD(w) DIV 10000H));
    WriteWord(f, WORD(LONGCARD(w) MOD 10000H));
  END WriteLong;

  PROCEDURE ReadLine(VAR f: File; VAR str: ARRAY OF CHAR);
    VAR i: CARDINAL;
        ch: CHAR;
  BEGIN
    i := 0;
    ReadChar(f, ch);
    WHILE (f.res = done) & (ch # EOL) & (i <= HIGH(str)) DO
      str[i] := ch; INC(i);
      ReadChar(f, ch)
    END;
    IF i <= HIGH(str) THEN str[i] := 0C END
  END ReadLine;

  PROCEDURE ReadBlock(VAR f: File; blkAdr: ADDRESS; blkLen: LONGCARD);
    VAR p: POINTER TO BYTE;
        err: INTEGER;
  BEGIN
    WriteBuffer(f);
    WITH f.int^ DO
      p := blkAdr;
      WHILE (bp # be) & (f.res = done) DO
        ReadByte(f, p^);
        p := ADDADR(p, 1); DEC(blkLen)
      END;
      IF f.res = done THEN
        err := read(path, p, blkLen);
        IF err = -1 THEN
          f.res := notdone; f.err := err
        ELSIF err # INTEGER(blkLen) THEN
          INC(filePos, err);
          f.res := eom; f.eof := TRUE
        ELSE
          f.res := done; f.eof := FALSE
        END
      END
    END
  END ReadBlock;

  PROCEDURE WriteBlock(VAR f: File; blkAdr: ADDRESS; blkLen: LONGCARD);
    VAR written: LONGCARD;
        err: INTEGER;
  BEGIN
    WriteBuffer(f);
    IF f.res = done THEN
      WITH f.int^ DO
        err := write(path, blkAdr, blkLen);
        IF err = -1 THEN
          f.res := notdone; f.err := err
        ELSIF err # INTEGER(blkLen) THEN
          INC(filePos, written);
          f.res := eom; f.eof := TRUE
        ELSE
          INC(filePos, written);
          f.res := done; f.eof := FALSE
        END
      END
    END
  END WriteBlock;

  PROCEDURE WriteRecord(VAR f: File; VAR record: ARRAY OF BYTE);
  BEGIN
    WriteBlock(f, ADR(record), SIZE(record))
  END WriteRecord;

  PROCEDURE ReadRecord(VAR f: File; VAR record: ARRAY OF BYTE);
  BEGIN
    ReadBlock(f, ADR(record), SIZE(record))
  END ReadRecord;

  PROCEDURE GetPath(VAR f: File): CARDINAL;
  BEGIN
    RETURN f.int^.path
  END GetPath;

  PROCEDURE GetError(VAR f: File): CARDINAL;
  BEGIN
    RETURN f.err
  END GetError;

END FileSystem.
