/* Copyright (C) 1989, Rowley Associates
                       32 Rowley
                       Cam
                       Dursley
                       Glos. GL11 5NT
                       England.

   RcsId: $Id: m2_start.c,v 1.5 93/01/21 21:48:03 paul Rel $

   Notes: This is the Modula-2 run-time support package programmed
          in 3L's Parallel C, Meiko's C, or the Perihelion C compiler.

          The functions in this system which take floats as parameters
          are not straight-coded in C because the Modula-2 compiler uses
          a different calling sequence from that of the C compilers.
          The Rowley Modula-2 compiler passes a 32-bit float as a 32-bit
          float unlike K&R C which passes a 32-bit float as a 64-bit
          double.  To overcome this, float parameters are defined as
          unsigneds and are converted to floats by assignment to and
          extraction from a union.

          *** 3L VARIANT ***
          Always compile with Parallel C's /S compiler option.
          Compile with /DSTANDALONE for a stand-alone run-time system.

          *** MEIKO VARIANT ***
          Always compile with Meiko C's -tK compiler option.
          Never compile with -DSTANDALONE.

          *** HELIOS VARIANT ***
          Always compile with stack checks off.
          Never compile with -DSTANDALONE.

*/

#include "m2rtl.h"

#ifdef _IMST2
#define STANDALONE
#endif

#ifndef STANDALONE
#include <stdio.h>
#endif

#include <string.h>
#include <setjmp.h>

/* Error handler. */
void (*M2RTL_ErrorHandler)(int);

/********** 3L VARIANT **********/
#ifdef _3L
#include <chan.h>
#endif

/********** HELIOS VARIANT **********/
#ifdef helios
#include <nonansi.h>
#include <stdlib.h>
/*
#include <syslib.h>
#include <stdlib.h>
#include <posix.h>
*/
extern char **environ;
#endif

#ifdef _3L
#ifdef _IMST2
CHAN *M2RTL_BootLinkIn, *M2RTL_BootLinkOut;
#else
int M2RTL_inlen, M2RTL_outlen;
CHAN *M2RTL_ins, *M2RTL_outs;
char **M2RTL_argv, **M2RTL_envp;
int M2RTL_argc;
#endif
#else
char **argv, **envp;
int argc;
#endif

static void WriteLn()
{
    char s[2];
    int n;

    s[0] = '\n'; n = 1;

#ifdef helios
    {
        Stream *strm = fdstream(STDERR);
        if (Flags_MSdos & strm->Flags)
        {
            s[0] = '\r'; s[1] = '\n'; n = 2;
        }
    }
#endif

#ifndef STANDALONE
    write(STDERR, s, n);
#endif
}


static void default_handler(code)
int code;
{
    char *str;

#ifdef STANDALONE
    /* stop this process */
    asm {stopp;}
#else

    switch (code)
    {
        case M2_CASE_ERROR:
          str = "No matching label in CASE statement"; break;
        case M2_IEEE_OVERFLOW:
          str = "IEEE overflow"; break;
        case M2_RETURN_ERROR:
          str = "No RETURN from function procedure"; break;
        case M2_STACK_OVERFLOW:
          str = "Stack overflow error"; break;
        case M2_NIL_DEREFERENCE:
          str = "NIL dereference"; break;
        case M2_RANGE_ERROR:
          str = "Value out of range"; break;
        case M2_NO_MEMORY:
          str = "No more memory (ALLOCATE)"; break;
        case M2_BAD_DISPOSE:
          str = "Double dispose of pointer (DEALLOCATE)"; break;
        case M2_PROCESS_TERM:
          str = "Process ended without TRANSFER"; break;
        case M2_DEADLOCK:
          str = "Deadlock in Processes - all processes waiting"; break;
        default:
          str = NULL;
    };

    if (str)
    {
        WriteLn();
        write(STDERR, "MODULA-2 RUN-TIME ERROR: ", 25);
        write(STDERR, str, strlen(str));
        WriteLn();
    }
    exit(code);
#endif
}


/********** 3L VARIANT **********/
#ifdef _3L

/* Main startup for 3L variant */

typedef struct
{
#ifdef _IMST2
  CHAN *boot_link_in;
  CHAN *boot_link_out;
#else
  int argc;
  char **argv;
  char **envp;
  CHAN **ins;
  int inlen;
  CHAN **outs;
  int outlen;
#endif
} ArgPack;

void M2RTL_START(ArgPack *args)
{
    /* Initialise M2RTL exported variables */

#ifdef _IMST2
    M2RTL_BootLinkIn = args->boot_link_in;
    M2RTL_BootLinkOut = args->boot_link_out;
#else
    M2RTL_argc = args->argc;
    M2RTL_argv = args->argv;
    M2RTL_envp = args->envp;
    M2RTL_ins = args->ins;
    M2RTL_inlen = args->inlen;
    M2RTL_outs = args->outs;
    M2RTL_outlen = args->outlen;
#endif

    M2RTL_ErrorHandler = default_handler;
}

/********** MEIKO & HELIOS VARIANT **********/
#else

/* Main startup for MeikOS, CS Tools, or Helios variant */
typedef struct
{
  int argc;
  char **argv;
  char **envp;
} ArgPack;

void M2RTL_START(ArgPack *args)
{
    /* Initialise M2RTL exported variables */
    argc = args->argc;
    argv = args->argv;

#ifdef helios
    envp = environ;
#else
    envp = args->envp;
#endif

    M2RTL_ErrorHandler = default_handler;
}

#endif
