IMPLEMENTATION MODULE MathLib0;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     RcsId: $Id: mathlib0.mod,v 1.5 92/07/25 17:36:02 paul Rel Locker: paul $

  *)

  (*V1=FAST    For processors with maths capability    *)
  (*V2=NOFAST  For processors without maths capability *)

  FROM RealConversions IMPORT
    GetRealExponent, SetRealExponent;
(*<FAST
  FROM M2RTL IMPORT
    MathChip, MathChipType;
  IMPORT
    FastMath;
>*)

  CONST
    piBy2 = pi / 2.0;
    piBy4 = pi / 4.0;
    sqrte = 1.64872127070013;  (* e**0.5 *)
    ln10  = 2.302585092994046; (* ln(10) *)

  TYPE
    Coeff11 = ARRAY [0..10] OF REAL;
    Coeff10 = ARRAY [0..9] OF REAL;
    Coeff8  = ARRAY [0..7] OF REAL;
    Coeff7  = ARRAY [0..6] OF REAL;
    Coeff6  = ARRAY [0..5] OF REAL;
    Coeff5  = ARRAY [0..4] OF REAL;
    Coeff4  = ARRAY [0..3] OF REAL;

  CONST
    LogCoeff = Coeff7{+1.0,
                      -1.0 / 2.0,
                      +1.0 / 3.0,
                      -1.0 / 4.0,
                      +1.0 / 5.0,
                      -1.0 / 6.0,
                      +1.0 / 7.0};

  CONST
    ExpCoeff = Coeff8{+1.00000000000000E+000,
                      +1.00000000000000E+000,
                      +5.00000000000000E-001,
                      +1.66666666666667E-001,
                      +4.16666666666667E-002,
                      +8.33333333333333E-003,
                      +1.38888888888889E-003,
                      +1.98412698412698E-004};

  CONST
    SinCoeff = Coeff5{+1.00000000000000E+000,
                      -1.66666666666667E-001,
                      +8.33333333333333E-003,
                      -1.98412698412698E-004,
                      +2.75573192239859E-006};

  CONST
    CosCoeff = Coeff5{+1.00000000000000E+000,
                      -5.00000000000000E-001,
                      +4.16666666666667E-002,
                      -1.38888888888889E-003,
                      +2.48015873015873E-005};


  CONST
    LngammaCoeff = Coeff6{+7.618009173E+001,
                          -8.650532033E+001,
                          +2.401409822E+001,
                          -1.231739515E+000,
                          +1.208580030E-003,
                          -5.363820000E-006};

  CONST
    ArctanCoeff = Coeff4{+1.0,
                         -1.0 / 3.0,
                         +1.0 / 5.0,
                         -1.0 / 7.0};

  CONST
    (* arctan(index/10.0) *)
    ArctanTable = Coeff11{+0.0,
                          +9.96686524911620E-002,
                          +1.97395559849881E-001,
                          +2.91456794477867E-001,
                          +3.80506377112365E-001,
                          +4.63647609000806E-001,
                          +5.40419500270584E-001,
                          +6.10725964389208E-001,
                          +6.74740942223552E-001,
                          +7.32815101786506E-001,
                          +piBy4};

  CONST
    LnTable = Coeff10{+0.000000000000000,
                      +0.095310179804325,
                      +0.182321556793955,
                      +0.262364264467491,
                      +0.336472236621213,
                      +0.405465108108164,
                      +0.470003629245735,
                      +0.530628251062170,
                      +0.587786664902119,
                      +0.641853886172394};

  (*$V-*)
  PROCEDURE ChebEval(x: REAL; C: ARRAY OF REAL): REAL;
    VAR i: CARDINAL;
        y: REAL;
  BEGIN
    y := C[HIGH(C)];
    FOR i := HIGH(C)-1 TO 0 BY -1 DO
      y := y * x + C[i];
    END;
    RETURN y
  END ChebEval;
  (*$V=*)

  PROCEDURE real(x: INTEGER): REAL;
  BEGIN
    IF x < 0 THEN RETURN -FLOAT(-x)
    ELSE RETURN FLOAT(x)
    END
  END real;

  PROCEDURE reald(x: LONGINT): REAL;
  BEGIN
    IF x < 0 THEN RETURN -FLOAT(-x)
    ELSE RETURN FLOAT(x)
    END
  END reald;

  PROCEDURE entier(x: REAL): INTEGER;
  BEGIN
    RETURN VAL(INTEGER, entierd(x))
  END entier;

  PROCEDURE entierd(x: REAL): LONGINT;
  BEGIN
    RETURN floor(x)
  END entierd;

  PROCEDURE truncate(x: REAL): LONGINT;
  BEGIN
    (* truncate towards zero. *)
    RETURN VAL(LONGINT, x)
  END truncate;

  PROCEDURE floor(x: REAL): LONGINT;
    VAR d: LONGINT;
  BEGIN
    (* truncate towards zero. *)
    d := VAL(LONGINT, x);

    (* adjust for truncation towards -infinity. *)
    IF (x < 0.0) & (VAL(REAL, d) # x) THEN
      DEC(d)
    END;

    (* Return result. *)
    RETURN d
  END floor;

  PROCEDURE ceiling(x: REAL): LONGINT;
    VAR d: LONGINT;
  BEGIN
    (* truncate towards zero. *)
    d := VAL(LONGINT, x);

    (* adjust for truncation towards +infinity. *)
    IF (x > 0.0) & (VAL(REAL, d) # x) THEN
      INC(d)
    END;

    (* Return result. *)
    RETURN d
  END ceiling;

  PROCEDURE round(x: REAL): LONGINT;
    VAR q: LONGINT;
  BEGIN
    (* round in appropriate direction. *)
    IF x >= 0.0 THEN
      q := VAL(LONGINT, x+0.5)
    ELSE
      q := VAL(LONGINT, x-0.5)
    END;

    (* Get fractional part. *)
    x := x - VAL(REAL, q);

    (* Correct rounding when number is exactly halfway
       between two integers. *)
    IF (x = 0.5) & ODD(q) THEN
      INC(q)
    END;
    IF (x = -0.5) & ODD(q) THEN
      DEC(q)
    END;

    (* Return result. *)
    RETURN q
  END round;

  PROCEDURE ffloor(x: REAL): REAL;
  BEGIN
    RETURN VAL(REAL, floor(x))
  END ffloor;

  PROCEDURE ftruncate(x: REAL): REAL;
  BEGIN
    RETURN VAL(REAL, truncate(x))
  END ftruncate;

  PROCEDURE fceiling(x: REAL): REAL;
  BEGIN
    RETURN VAL(REAL, ceiling(x))
  END fceiling;

  PROCEDURE fround(x: REAL): REAL;
  BEGIN
    RETURN VAL(REAL, fround(x))
  END fround;

  PROCEDURE sign(x: REAL): REAL;
  BEGIN
    IF x < 0.0 THEN    RETURN -1.0
    ELSIF x = 0.0 THEN RETURN  0.0
    ELSE               RETURN +1.0
    END
  END sign;

  PROCEDURE degtorad(angle: REAL): REAL;
  BEGIN
    RETURN angle * (pi / 180.0)
  END degtorad;

  PROCEDURE radtodeg(angle: REAL): REAL;
  BEGIN
    RETURN angle * (180.0 / pi)
  END radtodeg;

  PROCEDURE degtograd(angle: REAL): REAL;
  BEGIN
    RETURN angle * (100.0 / 90.0)
  END degtograd;

  PROCEDURE gradtodeg(angle: REAL): REAL;
  BEGIN
    RETURN angle * (90.0 / 100.0)
  END gradtodeg;

  PROCEDURE radtograd(angle: REAL): REAL;
  BEGIN
    RETURN angle * (200.0 / pi)
  END radtograd;

  PROCEDURE gradtorad(angle: REAL): REAL;
  BEGIN
    RETURN angle * (pi / 200.0)
  END gradtorad;

  PROCEDURE sqrt(x: REAL): REAL;
    VAR w: REAL;
        exp0, exp1: INTEGER;
        i: CARDINAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.sqrt(x)
    END;
>*)

    IF x >= 0.0 THEN
      GetRealExponent(x, exp0);
      IF ODD(exp0) THEN
        SetRealExponent(x, 7FH)
      ELSE
        SetRealExponent(x, 80H); DEC(exp0)
      END;

      (* Heron iterations: *)
      w := 1.0 + (x - 1.0) / 3.0;
      FOR i := 1 TO 3 DO
        w := (x/w + w) * 0.5;
      END;

      (* Insert correct exponent *)
      GetRealExponent(w, exp1);
      SetRealExponent(w, exp1 + (exp0-7FH) DIV 2);
      RETURN w
    ELSE
      HALT
    END;
  END sqrt;

  PROCEDURE sin(x: REAL): REAL;
    VAR octant: LONGINT;
        y: REAL;
        negArg: BOOLEAN;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.sin(x)
    END;
>*)

    (* apply tansform: sin(x) = -sin(-x), for x < 0 *)
    negArg := (x < 0.0);
    IF negArg THEN x := -x END;

    (* reduce x to primary domain. *)
    octant := LTRUNC(x / piBy4);
    x := x - FLOAT(octant) * piBy4;

    IF ODD(octant) THEN
      (* reflect in Y axis about X=pi/4 *)
      x := piBy4 - x
    END;

    (* reduce octant to primary domain. *)
    octant := octant MOD 8;

    (* apply polynomial. *)
    IF (octant+1) MOD 4 < 2 THEN
      (* for octants 0, 3, 4, 7 *)
      y := x * ChebEval(x*x, SinCoeff)
    ELSE
      (* for octants 1, 2, 5, 6 *)
      y := ChebEval(x*x, CosCoeff);
    END;

    IF (octant >= 4) # negArg THEN
      (* reflect in X axis about Y=0 *)
      y := -y;
    END;
    RETURN y
  END sin;

  PROCEDURE cos(x: REAL): REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.cos(x)
    END;
>*)

    RETURN sin(x + piBy2);
  END cos;

  PROCEDURE tan(x: REAL): REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.tan(x)
    END;
>*)

    RETURN sin(x) / cos(x)
  END tan;

  PROCEDURE sec(x: REAL): REAL;
  BEGIN
    RETURN 1.0 / cos(x)
  END sec;

  PROCEDURE csc(x: REAL): REAL;
  BEGIN
    RETURN 1.0 / sin(x)
  END csc;

  PROCEDURE cot(x: REAL): REAL;
  BEGIN
    RETURN cos(x) / sin(x)
  END cot;

  PROCEDURE ln(x: REAL): REAL;
    VAR y: REAL;
        exp: INTEGER;
        n: CARDINAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.ln(x)
    END;
>*)

    IF x > 0.0 THEN

      (* Reduce x to the range [1, 2] by setting the exponent to zero.     *)
      (* The exponent in REAL numbers is excess 127, hence 7FH below. *)
      GetRealExponent(x, exp);
      DEC(exp, 7FH);
      SetRealExponent(x, 7FH);

      (* reduce x to a number < 1.1 *)
      n := TRUNC((x-1.0) * 10.0);            (* n is tenths of x    *)
      x := x / (1.0 + FLOAT(n)/10.0) - 1.0;  (* apply log transform *)

      (* Use a Chebyshev polynomial to evaluate the logarithm on the
         transformed value and add the log of the exponent *)
      RETURN LnTable[n] + x*ChebEval(x, LogCoeff) + real(exp)*ln2;
    ELSE
      HALT
    END;
  END ln;

  PROCEDURE log(x: REAL): REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.log(x)
    END;
>*)

    RETURN ln(x) / ln10
  END log;

  PROCEDURE exp(x: REAL): REAL;
    VAR y, p: REAL;
        i: LONGINT;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.exp(x)
    END;
>*)

    (* e**(-x) = 1/(e**x) *)
    IF x < 0.0 THEN
      RETURN 1.0 / exp(-x)
    ELSE

      (* split x into an integer and fractional part *)
      i := LTRUNC(x);       (* integer part    *)
      x := x - FLOAT(i);    (* fractional part *)

      (* The polynomial approximation is good over the range [0, 0.5], so *)
      (* use that for x <= 0.5. For x > 0.5, we transform the exponental  *)
      (* e**(x) into e**(x-0.5) * e**(0.5). Thus, x-0.5 is reduced to the *)
      (* range [0, 0.5] for which the polynomial is a good approximation. *)
      IF x <= 0.5 THEN
        y := ChebEval(x, ExpCoeff)
      ELSE
        y := ChebEval(x - 0.5, ExpCoeff) * sqrte
      END;

      (* add in the integer part *)
      RETURN y * poweri(e, i)
    END
  END exp;

  PROCEDURE arctan(x: REAL): REAL;
    VAR y: REAL;
        n: CARDINAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.arctan(x)
    END;
>*)

    (* check for negative or greater than 1 *)
    IF x < 0.0 THEN
      RETURN -arctan(-x)
    ELSIF x > 1.0 THEN
      RETURN piBy2 - arctan(1.0 / x)
    ELSE
      (* range reduce to to a number between 0 and 0.1 by using the *)
      (* identity: atan(y) = atan(y) + atan((x - y) / (1 + x*y)     *)
      n := TRUNC(x * 10.0);
      IF n > 0 THEN
        y := FLOAT(n) / 10.0;
        x := (x-y) / (1.0 + x*y);
      END;

      RETURN ArctanTable[n] + x * ChebEval(x*x, ArctanCoeff);
    END
  END arctan;

  PROCEDURE arctan2(y, x: REAL): REAL;
    VAR primary: REAL;
  BEGIN
    IF (x = 0.0) & (y = 0.0) THEN

      (* x = 0, y = 0: Cartesian locus is the origin.
         This is really an error, but we return 0.0 instead. *)
      primary := 0.0

    ELSIF x = 0.0 THEN

      (* Cartesian locus somewhere in y-axis. *)
      primary := pi / 2.0

    ELSIF y = 0.0 THEN

      (* Cartesian locus somewhere in x-axis. *)
      primary := 0.0

    ELSE

      (* Compute result in Quadrant I. *)
      primary := arctan(ABS(y) / ABS(x));
    END;


    (* Compute results for Cartesian locus.

        Condition     Cartesian Locus      Range of result
        ---------     ---------------      ---------------

       y = 0, x > 0   Positive x-axis      0
       y > 0, x > 0   Quadrant I.          0 < result < pi/2
       y > 0, x = 0   Positive y-axis      pi/2
       y > 0, x < 0   Quadrant II.         pi/2 < result < pi
       y = 0, x < 0   Negative x-axis      pi
       y < 0, x < 0   Quadrant III.        -pi < result < -pi/2
       y < 0, x = 0   Negative y-axis      -pi/2
       y < 0, x > 0   Quadrant IV.         -pi/2 < result < 0
       y = 0, x = 0   Error                Defined to be 0

    *)

    IF x < 0.0 THEN
      primary := pi - primary
    END;
    IF y < 0.0 THEN
      primary := -primary
    END;

    RETURN primary
  END arctan2;

  PROCEDURE power(x, y: REAL): REAL;
    VAR r: REAL;
        n: LONGINT;
  BEGIN
    IF y < 0.0 THEN
      n := -VAL(LONGINT, LTRUNC(ABS(y)));
      r := y + FLOAT(ABS(n))
    ELSE
      n := LTRUNC(y);
      r := y - FLOAT(n)
    END;

    IF r = 0.0 THEN
      RETURN poweri(x, n)
    ELSE
      RETURN exp(y * ln(x))
    END
  END power;

  PROCEDURE poweri(x: REAL; n: LONGINT): REAL;
    VAR e: CARDINAL;
        xNeg: BOOLEAN;
        w, y: REAL;
  BEGIN
    IF n = 0 THEN RETURN 1.0       (* x ** 0 = 1.0 *)
    ELSIF x = 0.0 THEN
      IF n > 0 THEN RETURN 0.0     (* 0.0 ** x = 0.0, x > 0 *)
      ELSE HALT                    (* 0.0 ** x, x < 0 => undefined *)
      END
    END;

    (* save sign of argument and make argument positive *)
    xNeg := x < 0.0;
    x := ABS(x);

    (* x**(-n) = (1/x)**(-n) *)
    IF n < 0 THEN
      x := 1.0 / x;
      n := -n
    END;

    (* Least significant bit of the power defines whether it is an    *)
    (* even or odd power. Even powers always return positive results. *)
    IF ODD(n) THEN
      y := x
    ELSE
      y := 1.0;
      xNeg := FALSE    (* even powers of numbers are always positive *)
    END;

    w := x;
    n := n DIV 2;
    WHILE n > 0 DO
      w := w * w;
      IF ODD(n) THEN y := y*w END;
      n := n DIV 2
    END;

    (* place correct sign *)
    IF xNeg THEN y := -y END;     (* odd power of negative number *)
    RETURN y
  END poweri;

  PROCEDURE lngamma(x: REAL): REAL;
    VAR y, ser, tmp: REAL;
        i: CARDINAL;
  BEGIN
    y := x - 1.0;
    tmp := y + 5.5;
    tmp := (y + 0.5) * ln(tmp) - tmp;
    ser := 1.0;
    FOR i := 1 TO 6 DO
      x := x + 1.0;
      ser := ser + LngammaCoeff[i] / y;
    END;
    RETURN tmp + ln(2.50662827465*ser)
  END lngamma;

  PROCEDURE arcsin(x: REAL): REAL;
  BEGIN
    IF x = 1.0 THEN RETURN piBy2
    ELSE RETURN arctan(x / sqrt(1.0 - x*x))
    END
  END arcsin;

  PROCEDURE arccos(x: REAL): REAL;
  BEGIN
    RETURN piBy2 - arcsin(x)
  END arccos;

  PROCEDURE arccot(x: REAL): REAL;
  BEGIN
    RETURN sign(x) * (piBy2 - arctan(ABS(x)))
  END arccot;

  PROCEDURE arcsec(x: REAL): REAL;
  BEGIN
    RETURN arccos(1.0 / x)
  END arcsec;

  PROCEDURE arccsc(x: REAL): REAL;
  BEGIN
    RETURN arcsin(1.0 / x)
  END arccsc;

  PROCEDURE sinh(x: REAL): REAL;
    VAR eX: REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.sinh(x)
    END;
>*)

    eX := exp(x);
    RETURN 0.5 * (eX - 1.0/eX);
  END sinh;

  PROCEDURE cosh(x: REAL): REAL;
    VAR eX: REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.cosh(x)
    END;
>*)

    eX := exp(x);
    RETURN 0.5 * (eX + 1.0/eX)
  END cosh;

  PROCEDURE tanh(x: REAL): REAL;
    VAR e2X: REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.tanh(x)
    END;
>*)

    e2X := exp(x + x);
    RETURN (e2X - 1.0) / (e2X + 1.0)
  END tanh;

  PROCEDURE sech(x: REAL): REAL;
    VAR eX: REAL;
  BEGIN
    eX := exp(x);
    RETURN 2.0 / (eX + 1.0/eX)
  END sech;

  PROCEDURE csch(x: REAL): REAL;
    VAR eX: REAL;
  BEGIN
    eX := exp(x);
    RETURN 2.0 / (eX - 1.0/eX)
  END csch;

  PROCEDURE coth(x: REAL): REAL;
    VAR eX, eMX: REAL;
  BEGIN
    eX := exp(x);
    eMX := 1.0 / eX;
    RETURN (eX + eMX) / (eX - eMX)
  END coth;

  PROCEDURE arcsinh(x: REAL): REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.arcsinh(x)
    END;
>*)

    RETURN ln(x + sqrt(x*x + 1.0))
  END arcsinh;

  PROCEDURE arccosh(x: REAL): REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.arccosh(x)
    END;
>*)

    RETURN ln(x + sqrt(x*x - 1.0))
  END arccosh;

  PROCEDURE arctanh(x: REAL): REAL;
  BEGIN
(*<FAST
    IF MathChip # none THEN
      RETURN FastMath.arctanh(x)
    END;
>*)

    RETURN 0.5 * (ln((1.0 + x) / (1.0 - x)))
  END arctanh;

  PROCEDURE arcsech(x: REAL): REAL;
  BEGIN
    RETURN ln((1.0 + sqrt(1.0 - x*x)) / x)
  END arcsech;

  PROCEDURE arccsch(x: REAL): REAL;
  BEGIN
    RETURN ln(1.0/x + sqrt(x*x + 1.) / ABS(x))
  END arccsch;

  PROCEDURE arccoth(x: REAL): REAL;
  BEGIN
    RETURN 0.5 * ln((x+1.0) / (x-1.0))
  END arccoth;

END MathLib0.
