IMPLEMENTATION MODULE RealConversions;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     RcsId: $Id: realconv.mod,v 1.4 91/10/02 10:26:15 paul Rel $

  *)

  FROM SYSTEM IMPORT
    LOC, WORD, CAST, LOCSPERWORD;

  TYPE
    LONGSET = SET OF [0..31];

  TYPE
    PowerArray = ARRAY [0..8] OF LONGREAL;
    RoundArray = ARRAY [3..17] OF LONGREAL;

  CONST
    power = PowerArray{1.E+001,
                       1.E+002,
                       1.E+004,
                       1.E+008,
                       1.E+016,
                       1.E+032,
                       1.E+064,
                       1.E+128,
                       1.E+256};

    round = RoundArray{5.E-002,
                       5.E-003,
                       5.E-004,
                       5.E-005,
                       5.E-006,
                       5.E-007,
                       5.E-008,
                       5.E-009,
                       5.E-010,
                       5.E-011,
                       5.E-012,
                       5.E-013,
                       5.E-014,
                       5.E-015,
                       5.E-016};

  VAR
    BigEndian: BOOLEAN;

  PROCEDURE GetByte(VAR x: ARRAY OF LOC; index: CARDINAL): CARDINAL;
  BEGIN
    RETURN ORD(CAST(CHAR, x[index]))
  END GetByte;

  PROCEDURE PutByte(VAR x: ARRAY OF LOC; index: CARDINAL; val: WORD);
  BEGIN
    IF BigEndian THEN
      x[index] := val[LOCSPERWORD-1]
    ELSE
      x[index] := val[0]
    END
  END PutByte;

  PROCEDURE GetLongExponent(x: LONGREAL; VAR exp: INTEGER);
    VAR c: CARDINAL;
  BEGIN
    IF BigEndian THEN
      c := GetByte(x, 0)*100H + GetByte(x, 1)
    ELSE
      c := GetByte(x, 7)*100H + GetByte(x, 6)
    END;

    exp := c DIV 10H MOD 800H
  END GetLongExponent;

  PROCEDURE SetLongExponent(VAR x: LONGREAL; exp: INTEGER);
    VAR c: CARDINAL;
  BEGIN
    IF BigEndian THEN
      c := GetByte(x, 0)*100H + GetByte(x, 1);
      c := CAST(CARDINAL, CAST(BITSET, c) * BITSET{15, 0..3} +
                          CAST(BITSET, exp * 10H));
      PutByte(x, 0, c DIV 100H);
      PutByte(x, 1, c MOD 100H);
    ELSE
      c := GetByte(x, 7)*100H + GetByte(x, 6);
      c := CAST(CARDINAL, CAST(BITSET, c) * BITSET{15, 0..3} +
                          CAST(BITSET, exp * 10H));
      PutByte(x, 7, c DIV 100H);
      PutByte(x, 6, c MOD 100H);
    END
  END SetLongExponent;

  PROCEDURE GetRealExponent(x: REAL; VAR exp: INTEGER);
    VAR cast: RECORD
                CASE : INTEGER OF
                  1: l: LONGCARD;
                | 2: r: REAL;
                END
              END;
  BEGIN
    cast.r := x;
    exp := VAL(CARDINAL, cast.l DIV 800000H MOD 100H)
  END GetRealExponent;

  PROCEDURE SetRealExponent(VAR x: REAL; exp: INTEGER);
    VAR cast: RECORD
                CASE : INTEGER OF
                  1: b: LONGSET;
                | 2: r: REAL;
                END
              END;
  BEGIN
    cast.r := x;
    cast.b := cast.b * LONGSET{31, 0..22} +
              CAST(LONGSET, VAL(LONGCARD, exp)*800000H);
    x := cast.r
  END SetRealExponent;

  PROCEDURE ten(e: CARDINAL): LONGREAL;
    VAR i: CARDINAL;
        t: LONGREAL;
  BEGIN
    i := 0; t := 1.0;
    WHILE e > 0 DO
      IF ODD(e) THEN t := power[i] * t END;
      e := e DIV 2;
      INC(i)
    END;
    RETURN t
  END ten;

  PROCEDURE DecodeReal(    r:    LONGREAL;
                       VAR mant: LONGREAL;
                       VAR sign: LONGREAL;
                       VAR exp:  INTEGER);
    VAR e: INTEGER;
  BEGIN
    IF r = 0.0 THEN
      sign := 0.0; exp := 0; mant := 0.0
    ELSE
      IF r > 0.0 THEN sign := 1.0 ELSE sign := -1.0; r := -r END;
      GetLongExponent(r, e);
      IF e > 3FFH THEN
        (* 77/256 is approximately log10(2) *)
        exp := VAL(CARDINAL, VAL(LONGCARD, e-3FFH)*77 DIV 256);
        r := r / ten(exp);
        IF r >= 1.0 THEN r := r / 10.0; INC(exp) END
      ELSE
        exp := VAL(CARDINAL, VAL(LONGCARD, 3FFH-e)*77 DIV 256);
        r := r * ten(exp); exp := -exp;
        IF r < 0.1 THEN r := r * 10.0; DEC(exp) END;
      END;
      IF r >= 1.0 THEN r := r / 10.0; INC(exp) END;
      mant := r * 10.0; DEC(exp)
    END
  END DecodeReal;

  PROCEDURE ConvertRealToStr(    r:    LONGREAL;
                             VAR s:    ARRAY OF CHAR;
                                 n:    CARDINAL;
                             VAR done: BOOLEAN);

    VAR
      e, p: CARDINAL;
      i: INTEGER;
      sgn: LONGREAL;

    PROCEDURE Char(c: CHAR);
    BEGIN
      s[p] := c; INC(p)
    END Char;

  BEGIN
    (* Must have at least nine characters. *)
    IF n < 9 THEN n := 9 END;

    (* Check for enough room. *)
    done := n < HIGH(s);
    IF done THEN

      (* Compute number of significant digits, including decimal point. *)
      DEC(n, 6);
      p := 0;

      (* Pad with spaces at the front. *)
      WHILE n > 17 DO
        Char(' '); DEC(n)
      END;

      (* Put in leading sign or space. *)
      IF r < 0.0 THEN Char('-'); r := ABS(r)
      ELSE Char(' ')
      END;

      (* Split real into mantissa, mantissa sign, and exponent. *)
      DecodeReal(r, r, sgn, i);

      (* Round number for display, and compensate for overflow. *)
      r := r + round[n];
      IF r >= 10.0 THEN r := r / 10.0; INC(i) END;

      (* Put in integer part (one digit) and decimal point. *)
      Char(CHR(ORD('0') + TRUNC(r))); DEC(n);
      r := r - LFLOAT(TRUNC(r));
      Char('.'); DEC(n);

      (* Put in the fractional part. *)
      REPEAT
        r := r * 10.0;
        Char(CHR(ORD('0') + TRUNC(r)));
        r := r - LFLOAT(TRUNC(r));
        DEC(n)
      UNTIL n = 0;

      (* Put in the exponent. *)
      Char('E');
      IF i < 0 THEN Char('-'); i := -i ELSE Char('+') END;
      Char(CHR(i DIV 100 + VAL(INTEGER, '0')));
      Char(CHR(i MOD 100 DIV 10 + VAL(INTEGER, '0')));
      Char(CHR(i MOD 10 + VAL(INTEGER, '0')));

      (* Terminate the string. *)
      s[p] := 0C;
    END
  END ConvertRealToStr;

  PROCEDURE ConvertStrToReal(VAR r:    LONGREAL;
                                 s:    ARRAY OF CHAR;
                             VAR done: BOOLEAN);
    VAR ch: CHAR;
        i, e: INTEGER;
        n, k: CARDINAL;
        x: LONGREAL;
        buf: ARRAY [0..39] OF CHAR;
        neg, negE: BOOLEAN;

    PROCEDURE Next;
    BEGIN
      IF n > HIGH(s) THEN ch := 0C
      ELSE ch := s[n]; INC(n)
      END
    END Next;

  BEGIN
    (* Get first character. *)
    n := 0;
    done := FALSE;
    Next;

    (* Handle leading mantissa sign. *)
    IF ch = '-' THEN
      neg := TRUE; Next
    ELSE
      neg := FALSE;
      IF ch = '+' THEN Next END
    END;

    IF ('0' <= ch) & (ch <= '9') OR (ch = '.') THEN
      e := 0; k := 0; x := 0.0;
      (* strip leading blanks *)
      WHILE ch = '0' DO Next END;

      (* Form integer part of number. *)
      WHILE ('0' <= ch) & (ch <= '9') DO

        (* Only fifteen digits are significant. *)
        IF k < 15 THEN
          x := 10.0*x + LFLOAT(ORD(ch)-ORD('0')); INC(k)
        ELSE
          INC(e)
        END;
        Next
      END;

      (* Read fraction if present. *)
      IF ch = '.' THEN
        Next;
        WHILE ('0' <= ch) & (ch <= '9') DO
          IF k < 15 THEN
            x := 10.0*x + LFLOAT(ORD(ch)-ORD('0'));
            INC(k); DEC(e)
          END;
          Next
        END
      END;

      (* Read scale factor if present. *)
      IF CAP(ch) = 'E' THEN
        Next; i := 0;
        IF ch = '-' THEN negE := TRUE; Next
        ELSE
          negE := FALSE;
          IF ch = '+' THEN Next END
        END;
        IF ('0' <= ch) & (ch <= '9') THEN
          i := ORD(ch) - ORD('0'); Next;
          WHILE ('0' <= ch) & (ch <= '9') DO
            i := i*10 + VAL(INTEGER, ORD(ch) - ORD('0')); Next
          END;
        ELSE
          done := FALSE; RETURN;
        END;
        IF negE THEN DEC(e, i) ELSE INC(e, i) END
      END;

      (* Check for proper termination. *)
      done := ch = 0C;
      IF done THEN
        IF e < 0 THEN x := x / ten(-e)
        ELSE x := x * ten(e)
        END;
        IF neg THEN r := -x ELSE r := x END
      END
    END
  END ConvertStrToReal;

  VAR
    W: WORD;

BEGIN
  W := WORD(1);
  BigEndian := CHAR(W[0]) = 0C;
END RealConversions.
