IMPLEMENTATION MODULE Storage;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     RcsId: $Id: storage.mod,v 1.2 93/01/26 22:09:27 paul Exp $

     Notes:  The heap is implemented as a linked list of free blocks.
             These blocks are ordered according to their addresses; free
             blocks with low addresses are placed at the front of the free
             list, and blocks with higher addresses are at the end of the
             free list.

  *)

  FROM SYSTEM IMPORT
    ADDRESS;
  IMPORT
    Heaps;

  CONST
    NORAM = 6;

  VAR
    Heap: Heaps.HeapPtr;

  PROCEDURE ExtendHeap(minSize: LONGCARD; VAR done: BOOLEAN);
  BEGIN
    Heaps.ExtendHeap(Heap, minSize, done)
  END ExtendHeap;

  PROCEDURE ALLOCATE(VAR adr: ADDRESS; amount: LONGCARD);
  BEGIN
    Allocate(adr, amount);
    IF adr = NIL THEN
      HALT(NORAM)
    END
  END ALLOCATE;

  PROCEDURE DEALLOCATE(VAR adr: ADDRESS; amount: LONGCARD);
  BEGIN
    Deallocate(adr, amount)
  END DEALLOCATE;

  PROCEDURE Allocate(VAR adr: ADDRESS; amount: LONGCARD);
  BEGIN
    Heaps.Allocate(Heap, adr, amount)
  END Allocate;

  PROCEDURE Deallocate(VAR adr: ADDRESS; amount: LONGCARD);
  BEGIN
    Heaps.Deallocate(Heap, adr, amount)
  END Deallocate;

  PROCEDURE Available(amount: LONGCARD): BOOLEAN;
  BEGIN
    RETURN Heaps.Available(Heap, amount)
  END Available;

  PROCEDURE TotalHeapAvail(): LONGCARD;
  BEGIN
    RETURN Heaps.TotalHeapAvail(Heap)
  END TotalHeapAvail;

BEGIN
  Heap := Heaps.HeapPtr(NIL)
END Storage.
