IMPLEMENTATION MODULE Strings;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     RcsId: $Id: strings.mod,v 1.7 93/01/15 22:31:22 paul Rel $

  *)

  PROCEDURE Min(i, j: CARDINAL): CARDINAL;
    (* return minimum of i and j *)
    VAR min: CARDINAL;
  BEGIN
    IF i < j THEN min := i
    ELSE min := j
    END;
    RETURN min
  END Min;

  (*$V-*)
  PROCEDURE Length(str: ARRAY OF CHAR): CARDINAL;
    VAR i: CARDINAL;
  BEGIN
    i := 0;
    WHILE (i <= HIGH(str)) & (str[i] # 0C) DO
      INC(i)
    END;
    RETURN i
  END Length;
  (*$V=*)

  (*$V-*)
  PROCEDURE Assign(    from: ARRAY OF CHAR;
                   VAR to:   ARRAY OF CHAR);
    VAR copy, i: CARDINAL;
  BEGIN
    (* initialise *)
    i := 0;
    copy := Min(HIGH(from), HIGH(to));

    (* copy characters across *)
    WHILE (i <= copy) & (from[i] # 0C) DO
      to[i] := from[i]; INC(i)
    END;

    (* terminate string properly *)
    IF i <= HIGH(to) THEN to[i] := 0C END
  END Assign;
  (*$V=*)

  (*$V-*)
  PROCEDURE Append(VAR s0: ARRAY OF CHAR;
                       s1: ARRAY OF CHAR);
    VAR lens0, lens1, copy, i: CARDINAL;
  BEGIN
    (* initialise: calculate length of strings *)
    lens0 := Length(s0);
    lens1 := Length(s1);

    (* calculate number of characters to copy into s0 *)
    copy := Min(HIGH(s0)+1-lens0, lens1);

    (* append 'copy' characters onto s0 from s1 *)
    i := 0;
    WHILE i < copy DO
      s0[lens0] := s1[i];
      INC(lens0); INC(i)
    END;

    (* terminate string correctly *)
    IF lens0 <= HIGH(s0) THEN s0[lens0] := 0C END
  END Append;
  (*$V=*)

  (*$V-*)
  PROCEDURE Prepend(VAR s0: ARRAY OF CHAR;
                        s1: ARRAY OF CHAR);
  BEGIN
    Concat(s1, s0, s0)
  END Prepend;
  (*$V=*)

  PROCEDURE Insert(    substr: ARRAY OF CHAR;
                   VAR str:    ARRAY OF CHAR;
                       before: CARDINAL);
    (* NOTE: substr must be copied *)
    VAR lenstr, lensubstr, i: CARDINAL;
  BEGIN
    (* initialise: calculate string lengths *)
    lenstr := Length(str);
    lensubstr := Length(substr);

    (* adjust 'before' to length of 'str' if necessary *)
    before := Min(before, lenstr);

    (* create a space for the substring; moves 0C at end if necessary *)
    FOR i := Min(lenstr+lensubstr, HIGH(str)) TO before+lensubstr BY -1 DO
      str[i] := str[i-lensubstr]
    END;

    (* copy in the substring *)
    i := 0;
    WHILE (i < lensubstr) & (before+i <= HIGH(str)) DO
      str[before+i] := substr[i];
      INC(i)
    END
  END Insert;

  PROCEDURE Delete(VAR str:  ARRAY OF CHAR;
                       from: CARDINAL;
                       len:  CARDINAL);
    VAR strlen, i: CARDINAL;
  BEGIN
    (* initialise: calculate length of string and where to copy from *)
    strlen := Length(str);
    IF (from < strlen) & (len < strlen - from) THEN

      (* move back rest of string deleting the required part *)
      i := from + len;
      WHILE i < strlen DO
        str[from] := str[i];
        INC(from); INC(i)
      END;
    END;

    (* terminate string correctly *)
    IF from <= HIGH(str) THEN str[from] := 0C END
  END Delete;

  (*$V-*)
  PROCEDURE Replace(VAR str: ARRAY OF CHAR;
                        old: ARRAY OF CHAR;
                        new: ARRAY OF CHAR):CARDINAL;
    VAR pos: CARDINAL;
  BEGIN
    pos := Pos(old, str);
    IF pos <= HIGH(str) THEN
      Delete(str, pos, Length(old));
      Insert(new, str, pos)
    END;
    RETURN pos
  END Replace;
  (*$V=*)

  (*$V-*)
  PROCEDURE Copy(    str:    ARRAY OF CHAR;
                     from:   CARDINAL;
                     len:    CARDINAL;
                 VAR result: ARRAY OF CHAR);
    (* NOTE: str must be copied *)
    VAR i: CARDINAL;
  BEGIN
    len := Min(len, HIGH(result)+1);
    i := Length(str);
    IF (i <= from) OR (len = 0) THEN
      result[0] := 0C
    ELSE
      IF from+len > i THEN len := i-from END;
      FOR i := 0 TO len-1 DO result[i] := str[i+from] END;
      IF len <= HIGH(result) THEN result[len] := 0C END
    END
  END Copy;
  (*$V=*)

  PROCEDURE Concat(    str1:   ARRAY OF CHAR;
                       str2:   ARRAY OF CHAR;
                   VAR result: ARRAY OF CHAR);
  BEGIN
    Assign(str1, result);
    Append(result, str2);
  END Concat;

  PROCEDURE Reverse(VAR str: ARRAY OF CHAR);
    VAR i, len: CARDINAL;
        ch: CHAR;
  BEGIN
    len := Length(str);
    IF len > 0 THEN
      i := 0;
      DEC(len);
      WHILE i < len DO
        ch := str[i]; str[i] := str[len]; str[len] := ch;
        INC(i); DEC(len)
      END
    END
  END Reverse;

  (*$V-*)
  PROCEDURE PosFrom(substr: ARRAY OF CHAR;
                    str:    ARRAY OF CHAR;
                    from:   CARDINAL): CARDINAL;
    VAR i, k: INTEGER;
        j: CARDINAL;
  BEGIN
    k := Length(str) - Length(substr);
    i := from;
    WHILE i <= k DO
      j := 0;
      LOOP
        IF (j > HIGH(substr)) OR (substr[j] = 0C) THEN
          RETURN i
        END;
        IF substr[j] # str[i + VAL(INTEGER, j)] THEN EXIT END;
        INC(j)
      END;
      INC(i)
    END;
    RETURN HIGH(str)+1
  END PosFrom;
  (*$V=*)

  (*$V-*)
  PROCEDURE Pos(substr: ARRAY OF CHAR;
                str:    ARRAY OF CHAR): CARDINAL;
  BEGIN
    RETURN PosFrom(substr, str, 0)
  END Pos;
  (*$V=*)

  (*$V-*)
  PROCEDURE Index(ch: CHAR;
                  str: ARRAY OF CHAR): CARDINAL;
  BEGIN
    RETURN IndexFrom(ch, str, 0)
  END Index;
  (*$V=*)

  (*$V-*)
  PROCEDURE IndexFrom(ch:   CHAR;
                      str:  ARRAY OF CHAR;
                      from: CARDINAL): CARDINAL;
    VAR len: CARDINAL;
  BEGIN
    len := Length(str);
    LOOP
      IF from >= len THEN
        RETURN HIGH(str)+1
      END;
      IF str[from] = ch THEN
        RETURN from
      END;
      INC(from)
    END
  END IndexFrom;
  (*$V=*)

  (*$V-*)
  PROCEDURE PosLastFrom(substr: ARRAY OF CHAR;
                        str:    ARRAY OF CHAR;
                        from:   CARDINAL): CARDINAL;
    VAR pos, lastpos: CARDINAL;
  BEGIN
    from := Min(from, Length(str)-1);
    lastpos := HIGH(str)+1;
    pos := PosFrom(substr, str, 0);
    WHILE pos <= from DO
      lastpos := pos;
      pos := PosFrom(substr, str, pos+1)
    END;
    RETURN lastpos
  END PosLastFrom;
  (*$V=*)

  (*$V-*)
  PROCEDURE Compare(str1, str2: ARRAY OF CHAR): INTEGER;
    VAR high, i: CARDINAL;
        ch: CHAR;
  BEGIN
    (* Compute length of strings. *)
    IF HIGH(str1) > HIGH(str2) THEN
      high := HIGH(str2)
    ELSE
      high := HIGH(str1)
    END;

    i := 0;
    REPEAT
      ch := str1[i];
      IF ch # str2[i] THEN
        IF ch > str2[i] THEN
          RETURN 1
        ELSE
          RETURN -1
        END
      END;
      IF ch = 0C THEN
        RETURN 0
      END;
      INC(i)
    UNTIL i > high;
    IF HIGH(str1) = HIGH(str2) THEN
      RETURN 0
    END;
    IF HIGH(str1) > HIGH(str2) THEN
      IF str1[i] = 0C THEN
        RETURN 0
      ELSE
        RETURN 1
      END
    ELSE
      IF str2[i] = 0C THEN
        RETURN 0
      ELSE
        RETURN -1
      END
    END
  END Compare;
  (*$V=*)

  (*$V-*)
  PROCEDURE CharIndex(str: ARRAY OF CHAR; index: CARDINAL): CHAR;
    VAR ch: CHAR;
  BEGIN
    IF index < Length(str) THEN ch := str[index]
    ELSE ch := 0C
    END;
    RETURN ch
  END CharIndex;
  (*$V=*)

  PROCEDURE Capitalize(VAR str: ARRAY OF CHAR);
    VAR i: CARDINAL;
  BEGIN
    i := 0;
    WHILE (i <= HIGH(str)) & (str[i] # 0C) DO
      str[i] := CAP(str[i]);
      INC(i)
    END
  END Capitalize;

  PROCEDURE Uncapitalize(VAR str: ARRAY OF CHAR);
    VAR i: CARDINAL;
        ch: CHAR;
  BEGIN
    i := 0;
    WHILE (i <= HIGH(str)) & (str[i] # 0C) DO
      ch := str[i];
      IF ('A' <= ch) & (ch <= 'Z') THEN
        str[i] := CHR(ORD(ch) - ORD('A') + ORD('a'))
      END;
      INC(i)
    END
  END Uncapitalize;

END Strings.

